@@ -1,8 +1,11 @@
-define(["nbextensions/vpython_libraries/plotly.min",
+define(["base/js/utils",
+        "nbextensions/vpython_libraries/plotly.min",
         "nbextensions/vpython_libraries/glow.min",
-        "nbextensions/vpython_libraries/jquery-ui.custom.min"], function(Plotly) {
+        "nbextensions/vpython_libraries/jquery-ui.custom.min"], function(utils, Plotly) {
 
 var comm
+var ws = null
+var isopen = false
 
 window.Plotly = Plotly
 
@@ -16,7 +19,52 @@ IPython.notebook.kernel.comm_manager.register_target('glow',
         // Register handlers for later messages:
         comm.on_msg(onmessage);
         comm.on_close(function(msg) {console.log("glow comm channel closed")});
-    });
+		
+        if (msg.content.data.wsport !== undefined) {
+           // create websocket instance
+           var port = msg.content.data.wsport
+           var uri = msg.content.data.wsuri
+           var loc = document.location, new_uri, url;
+		
+           // Get base URL of current notebook server
+           var base_url = utils.get_body_data('baseUrl');
+           // Construct URL of our proxied service
+           var service_url = base_url + 'proxy/' + port + uri;
+
+           if (loc.protocol === "https:") {
+              new_uri = "wss:";
+           } else {
+              new_uri = "ws:";
+           }
+           if (document.location.hostname.includes("localhost")){
+              url = "ws://localhost:" + port + uri;
+           }
+           else {
+              new_uri += '//' + document.location.host + service_url;
+              url = new_uri
+           }
+           ws = new WebSocket(url);
+           ws.binaryType = "arraybuffer";
+           
+            // Handle incoming websocket message callback
+            ws.onmessage = function(evt) {
+                console.log("WebSocket Message Received: " + evt.data)
+            };
+ 
+            // Close Websocket callback
+            ws.onclose = function(evt) {
+                ws = null
+                isopen = false
+                console.log("***WebSocket Connection Closed***");
+            };
+ 
+            // Open Websocket callback
+            ws.onopen = function(evt) {
+                isopen = true
+                console.log("***WebSocket Connection Opened***");
+            };
+        }
+ });
 
 var datadir = requirejs.s.contexts._.config.paths.nbextensions + '/vpython_data/'
 window.Jupyter_VPython = datadir // prefix used by glow.min.js for textures
@@ -104,7 +152,12 @@ function send() { // periodically send events and update_canvas and request obje
 	var update = update_canvas()
 	if (update !== null) events = events.concat(update)
 	if (events.length === 0) events = [{event:'update_canvas', 'trigger':1}]
-	if (comm) comm.send( events )
+	if (ws && isopen) {
+		var msg = JSON.stringify(events)
+		ws.send(msg)
+	} else if (comm) {
+	     comm.send( events )
+    }
 	events = []
 }
 
@@ -132,6 +185,7 @@ var lastcenter = vec(0,0,0)
 var lastrange = 1
 var lastautoscale = true
 var lastsliders = {}
+var lastkeysdown = []
 var interval = 33 // milliseconds
 
 function update_canvas() { // mouse location and other stuff
@@ -148,6 +202,23 @@ function update_canvas() { // mouse location and other stuff
 		var ray = cvs.mouse.ray 
 		if (!ray.equals(lastray)) {evt.ray = [ray.x,ray.y,ray.z]; dosend=true}
 		lastray = ray
+
+		var k = keysdown()
+		var test = true // assume keysdown() is same as lastkeysdown
+		if (k.length !== lastkeysdown.length) test = false
+		else {
+			for (var i=0; i<k.length; i++) {
+				if (k[i] !== lastkeysdown[i]) {
+					test = false
+					break
+				}
+			}
+		}
+		if (!test) {
+			evt.keysdown = lastkeysdown = k
+			dosend = true
+		}
+
 		// forward and range may be changed by user (and up with touch), and autoscale (by zoom)
 		if (cvs.userspin) {
 			var forward = cvs.forward
@@ -305,7 +376,7 @@ function process_binding(event) { // event associated with a previous bind comma
 
 var sliders = {}
 
-function control_handler(obj) {  // button, menu, slider, radio, checkbox
+function control_handler(obj) {  // button, menu, slider, radio, checkbox, winput
     "use strict";
     var evt = {idx: obj.idx}
     if (obj.objName === 'button') {
@@ -327,6 +398,10 @@ function control_handler(obj) {  // button, menu, slider, radio, checkbox
     } else if (obj.objName === 'menu') {
         evt.value = obj.index
         evt.widget = 'menu'
+    } else if (obj.objName === 'winput') {
+        evt.text = obj.text
+        evt.value = obj.number
+        evt.widget = 'winput'
     } else {
         console.log('unrecognized control', 'obj=', obj, obj.text)
     }
@@ -368,26 +443,27 @@ var attrsb = {'a':'userzoom', 'b':'userspin', 'c':'range', 'd':'autoscale', 'e':
               'p':'left', 'q':'right', 'r':'top', 's':'bottom', 't':'_cloneid',
               'u':'logx', 'v':'logy', 'w':'dot', 'x':'dot_radius', 
               'y':'markers', 'z':'legend', 'A':'label','B':'delta', 'C':'marker_color',
-              'D':'size_units', 'E':'userpan'}
+              'D':'size_units', 'E':'userpan', 'F':'scroll', 'G':integrate_selected}
 
 // methods are X in {'m': '23X....'}
 var methods = {'a':'select', 'b':'pos', 'c':'start', 'd':'stop', 'f':'clear', // unused eghijklmnopvxyzCDFAB
 			   'q':'plot', 's':'add_to_trail',
                't':'follow', 'u':'_attach_arrow', 'w':'clear_trail',
                'G':'bind', 'H':'unbind', 'I':'waitfor', 'J':'pause', 'K':'pick', 'L':'GSprint',
-		       'M':'delete'}
+		       'M':'delete', 'N':'capture'}
          
 var vecattrs = ['pos', 'up', 'color', 'trail_color', 'axis', 'size', 'origin', '_attach_arrow',
                 'direction', 'linecolor', 'bumpaxis', 'dot_color', 'ambient', 'add_to_trail', 'textcolor',
                 'foreground', 'background', 'ray', 'ambient', 'center', 'forward', 'normal',
                 'marker_color']
                 
-var textattrs = ['text', 'align', 'caption', 'title', 'title_align', 'xtitle', 'ytitle', 'selected',
+var textattrs = ['text', 'align', 'caption', 'title', 'title_align', 'xtitle', 'ytitle', 'selected', 'capture',
                  'label', 'append_to_caption', 'append_to_title', 'bind', 'unbind', 'pause', 'GSprint']
 
 // patt gets idx and attr code; vpatt gets x,y,z of a vector            
 var patt = /(\d+)(.)(.*)/
 var vpatt = /([^,]*),([^,]*),(.*)/
+var quadpatt = /([^,]*),([^,]*),(.*)/
 var plotpatt = /([^,]*),([^,]*)/
 
 function decode(data) { 
@@ -416,6 +492,15 @@ function decode(data) {
 			if (vecattrs.indexOf(attr) > -1) {
 				val = m[3].match(vpatt)
 				val = vec(Number(val[1]), Number(val[2]), Number(val[3]))
+            } else if (attr == 'vs') {
+                var vs
+                val = m[3].match(quadpatt)
+                if (val === null) {
+                    val = m[3].match(vpatt)
+                    vs = [Number(val[1]), Number(val[2]), Number(val[3])]
+                } else {
+                    vs = [Number(val[1]), Number(val[2]), Number(val[3]), Number(val[4])]
+                }
 			} else if (textattrs.indexOf(attr) > -1) {
                 // '\n' doesn't survive JSON transmission, so in vpython.py we replace '\n' with '<br>'
 				val = m[3].replace(/<br>/g, "\n")
@@ -717,6 +802,13 @@ function handle_cmds(dcmds) {
 				cfg = fix_location(cfg)
 				glowObjs[idx] = wtext(cfg)
 				break
+            }
+            case 'winput': {
+				cfg.objName = obj
+				cfg.bind = control_handler
+				cfg = fix_location(cfg)
+				glowObjs[idx] = winput(cfg)
+				break
             }
 			case 'checkbox': {
 				cfg.objName = obj
@@ -845,6 +937,8 @@ function handle_methods(dmeth) {
 			obj.unbind(val, process_binding)
 		} else if (method === "follow") {
 			obj.camera.follow(glowObjs[val])
+		} else if (method === "capture") {
+			obj.capture(val)
 		} else if (method === 'waitfor') {
 			waitfor_canvas = idx
 			waitfor_options = val
@@ -907,6 +1001,9 @@ function handle_attrs(dattrs) {
 		} else {
 			if (triangle_quad.indexOf(attr) !== -1) {
 				obj[attr] = glowObjs[val]
+            } else if (attr == 'vs') {
+                if (val.length == 3) obj['vs'] = [ glowObjs[val[0]], glowObjs[val[1]], glowObjs[val[2]] ]
+                else obj['vs'] = [ glowObjs[val[0]], glowObjs[val[1]], glowObjs[val[2]], glowObjs[val[3]] ]
 			} else {
 				obj[attr] = val
 			}
@@ -915,4 +1012,4 @@ function handle_attrs(dattrs) {
 }
 console.log("END OF GLOWCOMM")
 
-});
\ No newline at end of file
+});
